/****************************************************************************
*****************************************************************************

          F R A M E    G R A B B E R    D E M O N S T R A T I O N

	Written   : K. J. Heaven
	Date      : 1 June 1988

*****************************************************************************
*****************************************************************************/

#define S_IREAD	 0x0100	/* from SYS\STAT.H */
#define S_IWRITE 0x0080	/* from SYS\STAT.H */

#define FREE_RUN  0x0033 /* bit pattern for free run mode */
#define HALT      0x0004 /* bit mask for halt, dont take next frame */
#define FREEZE    0x0079 /* bit pattern for freeze mode */

#define CLRREAD   0x00ed /* bit pattern for clear address register */
#define READ      0x00fd /* bit pattern for read in data */

#define CLRWRIT   0x00e7 /* bit pattern for clear before write */
#define WRITE     0x00f7 /* bit pattern for write mode */

#define READY     0x0080 /* bit pattern for grabber ready */

#define CONTROL   0x02E2 /* control port address */
#define DATAP     0x02E3 /* data port address */

typedef unsigned char PIXEL;
typedef unsigned int DISTRIBUTION;
typedef struct {
          int           length;			/* Length of chain */
	  unsigned int  start;			/* Start point offset */
	  int           closed;			/* Flag to indicate closed chain */
	  char          *element;               /* Pointer to where elements are */
	}  CHAIN;

#define NO_OF_CHAINS 50

CHAIN chains[NO_OF_CHAINS];                               /* Chain parameter work area */

char    BOXRJX = 182,		/* Box drawing characters */
        BOXHBX = 196,
	BOXLJX = 199,
	BOXRJ = 185,
	BOXVB = 186,
	BOXRT = 187,
       	BOXRB = 188,
       	BOXLB = 200,
        BOXLT = 201,
        BOXBJ = 202,
        BOXTJ = 203,
        BOXLJ = 204,
        BOXHB = 205,
        BOXCJ = 206;

#define BS 8
#define FORMFEED 12
#define CR 13
#define ESC 27
#define HOMEKEY 327
#define ENDKEY 335
#define UPKEY 328
#define DOWNKEY 336
#define PGUPKEY 329
#define PGDNKEY 337
#define LEFTKEY 331
#define INSKEY 338
#define RIGHTKEY 333
#define DELKEY 339
#define CTRLLEFTKEY 371
#define CTRLRIGHTKEY 372
#define F1 315
#define F2 316
#define F3 317
#define F4 318
#define F5 319
#define F6 320
#define F7 321
#define F8 322
#define F9 323
#define F10 324
#define TRUE 1
#define FALSE 0

#include <stdio.h>
#include <fcntl.h>
#include <alloc.h>
#include <stdlib.h>
#include <io.h>
#include <graphics.h>
#include <math.h>
#include <mem.h>
#include <ctype.h>
#include <bios.h>
#include <string.h>
#include <menu.h>
#include <dos.h>

/*     I/O Address 6/2E3 is used as the data port to the frame grabber
       Once configured, repeated accesses to this I/O location increment the
       internal address counter.

       I/O Address 6/2E2 is used as the control/status register.
       Bit patterns are as follows;

       Control outputs

        0 - nc
	1 - ram /r-w
        2 - /run
        3 - select
        4 - /clear address reg.
        5 - A to D /oe
	6 - nc
        7 - nc

       Status inputs

        7 - /ready
*/
/**************************************************************************/
/* Global data definitions, used by all routines */

int	running,	     	/* Current state of frame grabber */
        control,		/* Data last written to control port */
	graphic,		/* System in graphics mode */
	quadrant,               /* System in 4 quadrant mode */
	graphic_disp;           /* Graphic mode set */

PIXEL *buffer1_start;		/* Starting location of 64k frame buffer 1 */
PIXEL *buffer2_start;		/* Starting location of 64k frame buffer 2 */
PIXEL *quad_buf_start;		/* Starting location of 64k frame buffer 2 */
PIXEL *buffer9_start;		/* Starting location of 64k frame buffer 9
                                      for quadrant mode */

unsigned int nocursor = 0x2000;

int	screen_width = 79,
	screen_height = 24;
int     window_active = 1;

char file_name[26] = "";

#define MAX_CHAIN_BUFFER_SIZE 40000


/**************************************************************************/
/* Low level fuctions for frame grabber control  */


display_status( char *message )
  {
    int i;
    char modes[] = { 'S', 'Q' };
    char graphic_mode[] = { 'N', 'Y' };

    gotoxy( 3, screen_height - 1 );
    for( i = 3; i < screen_width; i++ )
      putchar( ' ' );
    gotoxy( 3, screen_height - 1 );  cprintf( message );
    gotoxy( 40, screen_height - 1 );
    i = coreleft() / 1024;
    cprintf( "%dk Free, Status %2x, Mode %c, Graphic %c",
			i,
			control,
			modes[quadrant],
			graphic_mode[graphic_disp] );
  }

outcont( int data_byte )		/* Writes to control register,        */
                        		/* and keeps ram copy up to date */
  {
    control = data_byte;
    outportb ( CONTROL, data_byte );
  }

halt ()				/* Halts the frame grabber at the end of */
  {				/* the current frame  */
    int i = 0;

    if ( running )  {
      outcont ( control | HALT );
      for ( i = 0; i < 10000 && !( 0x80 & inportb ( CONTROL )); i++ );
    }
    if ( i == 10000 )  {
      display_status ("Frame Grabber failed to come ready\7");
      delay ( 500 );
    }
    running = 0;
  }

free_run ()                     /* Sets the frame grabber in run mode */
  {
    outcont ( FREE_RUN );
    running = 1;
    display_status ("Frame grabber now in free run mode");
  }


freeze ()                       /* Holds the current frame being grabbed */
  {
    int i;

    halt ();			/* Inhibit next frame from starting */
    outcont ( FREEZE );         /* Set to re-display mode */
    running = 1;
    display_status ("Frame grabber now in freeze mode");
  }

load_buffer (PIXEL *buffer)
  {
    unsigned  i;

    halt ();        		/* Prepare frame grabber for read in */
    outcont ( CLRREAD );
    inportb ( DATAP );
    outcont ( READ );

    display_status ("Reading in from frame grabber ..");
    i = 65536;
    do  {
      *(buffer++) = inportb ( DATAP );
      i--;
    }
    while (i);
  }


store_buffer (PIXEL *buffer)
  {
    unsigned  i, j;
    char      *buff;
    char      *buffer_save;

    buffer_save = buffer;

    if (quadrant) {
      for ( i = 0; i < 128; i++ )
	memcpy ((buffer9_start + 256 * i) + 32768 + 128,
		(buffer9_start + 256 * i) + 32768, 128 );

      for ( i = 0; i < 128; i++ )
	memcpy ((buffer9_start + 256 * i) + 32768,
		(buffer9_start + 256 * i) + 128, 128 );

      for ( i = 0; i < 128; i++ )
	memcpy ((buffer9_start + 256 * i) + 128, buffer9_start + 256 * i, 128 );

      for ( i = 0; i < 128; i++ ) {
	for ( j = 0; j < 128; j++ ) {
	  *(buffer9_start + 256 * i + j) =
	    *(buffer + 512 * i + 256 + j * 2);
	}
      }
      buffer = buffer9_start;
      buffer_save = buffer9_start;
    }
    halt();

    outcont( CLRWRIT );	/* Prepare frame grabber for block write */
    inportb( DATAP );
    outcont( WRITE );
    display_status( "Writing to frame grabber ..");

    i = 65536;
    do  {
      outportb (DATAP, *(buffer++));
      i--;
    }
    while (i);

    if( graphic_disp ) {
      if ((buff = malloc ((screen_width + 1) * (screen_height + 1) * 2)) == NULL) {
        printf("\7\nUnable to allocate necessary memory");
        exit( 1 );
      }
      gettext( 1, 1, screen_width, screen_height, buff );
      display_frame( buffer_save );
      puttext( 1, 1, screen_width, screen_height, buff );
      free( buff );
      return( 0 );
    }
  }




/****************************************************************************/
/* Mid level functions for menu options */


int getkey(void)
/* Uses the BIOS to read the next keyboard character */
{
 int key, lo, hi;

 key = bioskey(0);
 lo = key & 0X00FF;
 hi = (key & 0XFF00) >> 8;
 return((lo == 0) ? hi + 256 : lo);
} /* getkey */


int editstring(char *s, char *legal, int maxlength, int x, int y)
/* Allows the user to edit a string with only certain characters allowed -
   Returns TRUE if ESC was not pressed, FALSE is ESC was pressed.
*/
{
 int c, len = strlen(s), pos = len, insert = TRUE;

 do
 {
  gotoxy (x, y );
  cprintf("%s", s);
  gotoxy(pos + x, y);
  putch (' ');
  gotoxy(pos + x, y);
  switch(c = getkey())
  {
   case HOMEKEY :
    pos = 0;
    break;
   case ENDKEY :
    pos = len;
    break;
   case INSKEY :
    insert = !insert;
    break;
   case LEFTKEY :
    if (pos > 0)
     pos--;
    break;
   case RIGHTKEY :
    if (pos < len)
     pos++;
    break;
   case BS :
    if (pos > 0)
    {
     movmem(&s[pos], &s[pos - 1], len - pos + 1);
     pos--;
     len--;
    }
    break;
   case DELKEY :
    if (pos < len)
    {
     movmem(&s[pos + 1], &s[pos], len - pos);
     len--;
    }
    break;
   case CR :
    break;
   case UPKEY :
    c = CR;
    break;
   case DOWNKEY :
    c = CR;
    break;
   case ESC :
    len = 0;
    break;
   default :
    if (((legal[0] == 0) || (strchr(legal, c) != NULL)) &&
        ((c >= ' ') && (c <= '~')) &&
        (len < maxlength))
    {
     if (insert)
     {
      memmove(&s[pos + 1], &s[pos], len - pos + 1);
      len++;
     }
     else if (pos >= len)
      len++;
     s[pos++] = c;
    }
    break;
  } /* switch */
  s[len] = 0;
 }
 while ((c != CR) && (c != ESC));
 return(c != ESC);
} /* editstring */

int get_file_name ( int x, int y, char *file_name )
  {
    int       i, j, width = 40, ret_code;
    char      *buff;

    if ((buff = malloc (width * 3 * 2)) == NULL) {
      printf("\7\nUnable to allocate necessary memory");
      exit(0);
    }
    gettext( x, y, x + width + 1, y + 2, buff );

    gotoxy( x, y );
    putch( BOXLT );
    for( j = 0; j <  width; j++ )
      putch( BOXHB );
    putch( BOXRT );

    gotoxy( x, y + 1 );
    putch( BOXVB );
    for( j = 0; j <  width; j++ )
      putch( ' ' );
    gotoxy( x + 2, y + 1 );
    highvideo();
    cprintf( "Filename : " );
    normvideo();
    gotoxy( x + width + 1, y + 1);
    putch( BOXVB );

    gotoxy( x, y + 2 );
    putch( BOXLB );
    for( j = 0; j <  width; j++ )
      putch( BOXHB );
    putch( BOXRB );

    ret_code = editstring(file_name,
	"01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_.:\\",
         25, x + 13, y + 1);

/*    gotoxy(x + 13, y + 1);
    cscanf("%25[ -~]", file_name ); */
    if ( !strchr ( file_name, '.'))
      strcat ( file_name, ".pic");

    puttext ( x, y, x +  width + 1, y + 2, buff );
    free ( buff);
    return (ret_code);
  }

buffer_to_disk (PIXEL *buffer, MENU *menu_params)    /* Transfers the frame from the grabber */
  {
    char   ch;
    int    work_file;
    int    i, j;
    char   work_area[60];

    strcpy( file_name, "");
    if ( get_file_name (menu_params -> x,
			menu_params -> y + menu_params -> opt + 1,
			file_name) ) {
      if ( !access ( file_name, 0))  {
	display_status( "Overwrite ? (y/n) " );
        gotoxy( 22, screen_height - 1 );
        cscanf( "%c", &ch);
        if ( 'N' == toupper ( ch ))
          return( 1 );
      }
      if (( work_file = open( file_name, O_RDWR | O_CREAT | O_TRUNC | O_BINARY,
                                    S_IREAD | S_IWRITE)) == -1) {
        gotoxy( 1, 24);
	sprintf( work_area, "Unable to open %s ", file_name );
	display_status( work_area );
        return( 1 );
      }
      display_status( "Saving ... ");

      halt();
      outcont( CLRREAD );
      inportb( DATAP );
      outcont( READ );

      for( i = 0; i < 16; ++i )  {
        for( j = 0; j < 4096; buffer[j++] = inportb( DATAP ));
          write( work_file, buffer, 4096);
      }
      freeze();
      close( work_file);
    }
  }

/*...........................................................................
	Transfer an image from disk to a buffer

	This routine prompts for a filename, then reads the data into the
	specified buffer.

     Required :	buffer		Buffer to use to store the image data
*/

disk_to_buffer( PIXEL *buffer, MENU *menu_params)
  {
    char   ch[2];
    int    work_file;
    int    i, j;
    char   work_area[60];

    if( get_file_name (menu_params -> x + menu_params -> width,
			menu_params -> y + menu_params -> opt + 1,
    			file_name) ) {
      if (( work_file = open( file_name, O_RDWR | O_BINARY )) == -1)  {
        sprintf( work_area, "Unable to open %s ", file_name );
	display_status( work_area );
        return( 1 );
      }
      for( i = 0; i < 16; i++, buffer += 4096 )
        read( work_file, buffer, 4096 );
      freeze();
      close( work_file );
    }
  }

clear_result_window()
  {
    if( !window_active ) {
      window( 1, 1, screen_width, screen_height + 1 );
      window_active = 1;
    }
  }

set_result_window()
  {
    if( window_active ) {
      window( 15, 17, screen_width - 3, screen_height - 3 );
      clrscr();
      window_active = 0;
    }
  }


test_memory()
  {
    char  ch, seed;
    long  i;
    int   error_count;
    char  work_area[60];

    halt();				/* Inhibit the next frame from starting */

    sprintf( work_area, "Writing, I/O base add. %X .. ", CONTROL );
    display_status( work_area );

    outcont ( CLRWRIT );
    inportb ( DATAP );
    outcont ( WRITE );

    for ( i = 0, seed = 0; i < 65536; outportb (DATAP, seed++), i++);

    display_status ("Reading ... ");
    error_count = 0;
    outcont ( CLRREAD );
    inportb ( DATAP );
    outcont ( READ );
    for ( i = 0; i < 10000; i++);
    for ( i = 0, seed = 0; i < 65536; i++, seed++ )  {
      ch = inportb ( DATAP );
      if ( seed != ch )  {
	set_result_window();
	gotoxy( 10, wherey() );
	cprintf( "\nError at %4lx should be %2x, is %2x", i, seed, ch );
	delay( 100 );
	if ( error_count++ > 20 )  {
	  cprintf( "\n\7Greater than 20 errors !!" );
	  cscanf ("%c", &ch);
          clear_result_window();
          freeze ();        /* Set to re-display mode */
	  return;
	}
      }
    }
    clear_result_window();
    freeze (); 			/* Set to re-display mode */
  }


histogram (PIXEL *in_buffer, PIXEL *out_buffer)
  {
    DISTRIBUTION distribution[64];

    calc_distribution (distribution, in_buffer);
    memcpy (out_buffer, in_buffer, 65535);
    overlay_distribution (distribution, 170, 20, out_buffer);
  }

horz_line (int startx, int starty, int length, PIXEL *buffer )
  {
    int i, j;
    for (i = starty * 256 + startx, j = 0; j < length; j++ )
      *(buffer + i++) = 255;
  }

vert_line (int startx, int starty, int length, PIXEL *buffer )
  {
    int i, j;
    for (i = starty * 256 + startx, j = 0; j < length; j++ )
      *(buffer + (i += 256)) = 255;
  }

overlay_distribution (DISTRIBUTION distribution[64], int x, int y,
                      PIXEL *buffer)
  {
    int i, j, k, intensity;

    display_status ("Overlaying distribution ..");


/* Draw a box outline, blanking out the histogram area */

    for ( i = y, j = 0; j < 50; i++, j++ )
      horz_line ( x, i, 66, buffer );
    horz_line ( x, y + 64, 66, buffer );
    vert_line ( x, y, 64, buffer );
    vert_line ( x + 65, y, 64, buffer );

/* Draw the grey scale along the bottom */

    for (i = 0; i < 64; i++)
      for (j = 1 + i + x + (y + 50) * 256, k = 0; k < 14; k++, j+= 256 )
	*(buffer + j) = i * 4;

/* Draw in the distribution over the blanked area */

    for (i = 0; i < 64; i++) {
      if (( intensity = 48 - distribution[i] / 128) < 0 )
	intensity = 0;
      for (j = 1 + x + i + ( (y + 1) * 256 ), k = 0; k < intensity; j += 256, k++ )
	 *(buffer + j) = 0;
    }
  }

calc_distribution (DISTRIBUTION distribution[64], PIXEL *buffer )
  {
    unsigned i;

    display_status ("Calculating distribution ..");

    for (i = 0; i < 64; distribution[i++] = 0);
    for (i = 0; i < 65535; i++ ) {
      distribution[ *(buffer + i) / 4 ]++;
    }
    distribution[ *(buffer + i) / 4]++;
  }

float calc_mean (PIXEL *buffer)
  {
    int i;
    long int mean;

    display_status ("Calculating mean ..");

    i = 65536;
    mean = 0;
    do {
      mean += *( buffer++ );
      i--;
    }
    while (i);
    return ( mean / 65536 );
  }

float calc_std_dev (PIXEL *buffer)
  {
    int i;
    unsigned long sumx, sumx2;
    double temp, temp1;

    display_status ("Calculating std. deviation ..");

    i = 65536;
    sumx = 0;
    sumx2 = 0;
    do {
      sumx += *( buffer );
      sumx2 += *( buffer ) * *( buffer++ );
      i--;
    }
    while (i);

    /* temp is used to force integer to float conversion etc. */

    temp = sumx;
    temp1 = sumx2;
    temp = sqrt ( abs ((( temp1 - temp * temp / 65536 ) / 65535 )));
    return ( temp );
  }

equalise (PIXEL *in_buffer, PIXEL *out_buffer)
  {
    float    mean, std_dev, scale_factor, shift_factor;
    int      temp;
    unsigned i;
    PIXEL    *trans_table;

    if ((trans_table = (PIXEL *) malloc (256 * sizeof (PIXEL))) == NULL) {
      printf ("\n\7Unable to allocate memory");
      exit(1);
    }

    mean = calc_mean (in_buffer);
    std_dev = calc_std_dev (in_buffer);

    set_result_window();
    cprintf("\nMean = %7.2f", mean);
    cprintf("\nStd_dev = %7.2f", std_dev);
    clear_result_window();

    scale_factor = 70 / ( std_dev + 1 );
    shift_factor = 128 - mean * scale_factor;

    for ( i = 0; i < 256; i++ ) {
      if (( temp = i * scale_factor + shift_factor ) < 0)
	*( trans_table + i ) = 0;
      else
	*( trans_table + i ) = temp;
        if ( temp > 255)
	  *( trans_table + i ) = 255;
    }

    i = 65536;
    do  {
      *( out_buffer++ ) = *( trans_table + *( in_buffer++ ));
      i--;
    }
    while (i);

    free (trans_table);
  }

display_frame (PIXEL *buffer)
  {
    int x, x_start, x_max,
	y, y_start, y_max,
	i, temp, dot, g_error,
	graph_width = 512,
	graph_height = 256;

    struct palettetype palette;

    int graphdriver = DETECT, graphmode;
    char ch;

    graphic = 1;

    initgraph( &graphdriver, &graphmode, "");
    if( graphdriver < 0 ) {
      display_status( grapherrormsg( g_error = graphresult() ) );
      }

    getpalette( &palette );
    dot = palette.size - 1;

    x_start = ( getmaxx() - graph_width ) / 2;
    y_start = ( getmaxy() - graph_height ) / 2;
    x_max = x_start + graph_width;
    y_max = y_start + graph_height;
    if( y_start < 0 )  {      		/* For CGA */
      y_start = 0;
      y_max = 200;
    }

    line( x_start, y_start, x_max, y_start );
    for( y = y_start; y < y_max; y += 2 ) {
      putpixel( x_start, y, dot );
      putpixel( x_start, y + 1, dot );
      for( x = x_start + 1; x < x_max; x += 4 )  {
	temp = *(buffer) + *(buffer + 1)
	     + *(buffer + 256) + *(buffer + 257 );
	buffer += 2;
	if( temp > 113 ) {
	  putpixel( x + 1, y, dot );
	  if( temp > 227 ) {
	    putpixel( x + 2, y + 1, dot);
	    if( temp > 340 ) {
	      putpixel( x + 3, y, dot);
	      if( temp > 450 ) {
		putpixel( x, y + 1, dot);
		if( temp > 567 ) {
		  putpixel( x, y, dot);
		  if( temp > 680 ) {
		    putpixel( x + 3, y + 1, dot);
		    if( temp > 793 ) {
		      putpixel( x + 1, y + 1, dot);
		        if( temp > 906 )
		        putpixel( x + 2, y, dot);
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
      putpixel( x_max, y, dot );
      putpixel( x_max, y + 1, dot );
      buffer += 256;
    }
    line( x_start, y_max, x_max, y_max );
    cscanf("%c", &ch );
    closegraph();
    graphic = 0;
  }



/*...........................................................................
    	Scan for a starting point to generate chain code from.

    Requires	: buffer	Buffer to scan
		: start		Position of last start point within buffer

    Returns	: 0		No more start points found
    		: 1		Good start point found
		: *start        Position of start point

*/

int scan_start_point (PIXEL *buffer, unsigned *start)
  {
    unsigned i;

    i = *start;
    buffer += *start;
    do {
      i++;
      buffer++;
    }
    while ((i < 65535) && ( *buffer == 0));

    if (i == 65535)		/* No more start points found */
      return(0);

    if (*buffer == 255) {	/* Set new start point */
      *start = i;
      return(1);
    }
  }

/*.........................................................................
        Apply Roberts Gradient edge detector to an image in a buffer

      Requires	: in_buffer	Input buffer to process
		: out_buffer	Output buffer to write edge detected to
*/

roberts_grad (PIXEL *in_buffer, PIXEL *out_buffer)
  {
    int      temp1, temp2;
    unsigned i;
    PIXEL    *out, *in;

    out = out_buffer;
    in  = in_buffer;

    display_status( "Calculating Roberts Gradient" );

    for( i = 0; i < 65280; i++, out++, in++) {
      temp1 = *(in + 257) - *in;
      temp2 = *(in + 1) - *(in + 256);
      *out = 255;
      if ((temp1 * temp1 + temp2 * temp2) < 512 )
	*out = 0;
    }
  }

/*...........................................................................
	Calculate the chain code for a chain, given a start point.

	This routine steps along an unbroken sequence of pixels from a
	start point, generating the chain code.

        Chain is coded as per the following map.

                  0  |  1  |  2
		-----+-----+-----
		  7  |  x  |  3
		-----+-----+-----
		  6  |  5  |  4

    The parameter restart, indicates that the routine should restart from
    the given point to attempt to overcome a false finish caused by noise.

    Requires :	start		position of start point in buffer specified
		chain		chain data area where chain to be written
		buffer          Buffer. Note that data gets destroyed
                restart         1 indicates that restart attempt is to be made

    The maximum length of a chain is 10000 steps.  Thus *chain needs to
    have 10000 free bytes.

    Returns  :  chain 		Chain data
*/

int calc_chain ( PIXEL *buffer,
                 CHAIN *chain,
		 int restart,
		 unsigned int *current_point)
  {
    int           i;
    char          ch;

/* Set up current position and zero length in case object not found.  This
   code is executed only for a first time start, not a restart.  */

    if (!restart) {
      *current_point = chain -> start;
      chain -> length = 0;

/* Calculate first direction.  Search from direction 6 through 3.
   Prior raster scan ensures direction can be none other.  While
   statement is used to allow use of break statement to exit. */

      while(1) {
        if ( *( buffer + (*current_point += 255)) > 128) {         /* Cen -> 6 */
          *(chain -> element) = 6;
          break;
        }
        if ( *( buffer + (*current_point += 1)) > 128) {           /* 6 - 5 */
          *(chain -> element) = 5;
          break;
        }
        if ( *( buffer + (*current_point += 1)) > 128) {           /* 5 - 4 */
          *(chain -> element) = 4;
          break;
        }
        if ( *( buffer + (*current_point -= 256)) > 128) {         /* 4 - 3 */
          *(chain -> element) = 3;
          break;
        }
      *(buffer + chain -> start) = 0;       /* Blank out the start point */
      return (chain -> closed = 0);         /* Can out if no pixel found */
      }
    }

/* Now we have the first direction, use this to determine which way to
   search for the next direction.  This is the start point when the restart
   parameter is active
*/

    for ( ; chain -> length < 10000; (chain -> length)++ ) {
      switch ( *(chain -> element + chain -> length))
	{
          case 0 :                                /* Last direction was 0, so search from 2 to 5 */
            while (1) {
              if ( *( buffer + (*current_point -= 255)) > 128) {         /* Cen -> 2 */
                *(chain -> element + chain -> length + 1) = 2;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {            /* 2 - 1 */
                *(chain -> element + chain -> length + 1) = 1;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {            /* 1 - 0 */
                *(chain -> element + chain -> length + 1) = 0;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {          /* 0 - 7 */
                *(chain -> element + chain -> length + 1) = 7;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {          /* 7 - 6 */
                *(chain -> element + chain -> length + 1) = 6;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {            /* 6 - 5 */
                *(chain -> element + chain -> length + 1) = 5;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;

          case 1 :                                /* Last direction was 1, so search from 2 to 6 */
            while (1) {
              if ( *( buffer + (*current_point -= 255)) > 128) {         /* Cen -> 2 */
                *(chain -> element + chain -> length + 1) = 2;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {            /* 2 - 1 */
                *(chain -> element + chain -> length + 1) = 1;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {            /* 1 - 0 */
                *(chain -> element + chain -> length + 1) = 0;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {          /* 0 - 7 */
                *(chain -> element + chain -> length + 1) = 7;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {          /* 7 - 6 */
                *(chain -> element + chain -> length + 1) = 6;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;

          case 2 :                                /* Last direction was 2, so search from 4 to 7 */
            while (1) {
	      if ( *( buffer + (*current_point += 257)) > 128) {         /* Cen -> 4 */
                *(chain -> element + chain -> length + 1) = 4;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 4 - 3 */
                *(chain -> element + chain -> length + 1) = 3;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 3 - 2 */
                *(chain -> element + chain -> length + 1) = 2;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {           /* 2 - 1 */
                *(chain -> element + chain -> length + 1) = 1;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {           /* 1 - 0 */
                *(chain -> element + chain -> length + 1) = 0;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {         /* 0 - 7 */
                *(chain -> element + chain -> length + 1) = 7;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;

          case 3 :                                /* Last direction was 3, so search from 4 to 0 */
            while (1) {
	      if ( *( buffer + (*current_point += 257)) > 128) {         /* Cen -> 4 */
                *(chain -> element + chain -> length + 1) = 4;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 4 - 3 */
                *(chain -> element + chain -> length + 1) = 3;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 3 - 2 */
                *(chain -> element + chain -> length + 1) = 2;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {           /* 2 - 1 */
                *(chain -> element + chain -> length + 1) = 1;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {           /* 1 - 0 */
                *(chain -> element + chain -> length + 1) = 0;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;

          case 4 :                                /* Last direction was 4, so search from 6 to 1 */
            while (1) {
	      if ( *( buffer + (*current_point += 255)) > 128) {         /* Cen -> 6 */
                *(chain -> element + chain -> length + 1) = 6;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 6 - 5 */
                *(chain -> element + chain -> length + 1) = 5;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 5 - 4 */
                *(chain -> element + chain -> length + 1) = 4;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 4 - 3 */
                *(chain -> element + chain -> length + 1) = 3;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 3 - 2 */
                *(chain -> element + chain -> length + 1) = 2;
                break;
              }
              if ( *( buffer + (*current_point -= 1)) > 128) {           /* 2 - 1 */
                *(chain -> element + chain -> length + 1) = 1;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;

          case 5 :                                /* Last direction was 5, so search from 6 to 2 */
            while (1) {
	      if ( *( buffer + (*current_point += 255)) > 128) {         /* Cen -> 6 */
                *(chain -> element + chain -> length + 1) = 6;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 6 - 5 */
                *(chain -> element + chain -> length + 1) = 5;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 5 - 4 */
                *(chain -> element + chain -> length + 1) = 4;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 4 - 3 */
                *(chain -> element + chain -> length + 1) = 3;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {         /* 3 - 2 */
                *(chain -> element + chain -> length + 1) = 2;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;

          case 6 :                                /* Last direction was 6, so search from 0 to 3 */
            while (1) {
	      if ( *( buffer + (*current_point -= 257)) > 128) {         /* Cen -> 0 */
                *(chain -> element + chain -> length + 1) = 0;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {         /* 0 - 7 */
                *(chain -> element + chain -> length + 1) = 7;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {         /* 7 - 6 */
                *(chain -> element + chain -> length + 1) = 6;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 6 - 5 */
                *(chain -> element + chain -> length + 1) = 5;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 5 - 4 */
                *(chain -> element + chain -> length + 1) = 4;
                break;
              }
              if ( *( buffer + (*current_point -= 256)) > 128) {           /* 4 - 3 */
                *(chain -> element + chain -> length + 1) = 3;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;

          case 7 :                                /* Last direction was 7, so search from 0 to 4 */
            while (1) {
	      if ( *( buffer + (*current_point -= 257)) > 128) {         /* Cen -> 0 */
                *(chain -> element + chain -> length + 1) = 0;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {         /* 0 - 7 */
                *(chain -> element + chain -> length + 1) = 7;
                break;
              }
              if ( *( buffer + (*current_point += 256)) > 128) {         /* 7 - 6 */
                *(chain -> element + chain -> length + 1) = 6;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 6 - 5 */
                *(chain -> element + chain -> length + 1) = 5;
                break;
              }
              if ( *( buffer + (*current_point += 1)) > 128) {           /* 5 - 4 */
                *(chain -> element + chain -> length + 1) = 4;
                break;
              }
            return (chain -> closed = 0);         /* Can out if no pixel found */
	  }
	  break;
	}

/*  We got here (didn't take a return() ), so a new chain point was found.
    Blank the point, so that it is not re-traced at a later stage.  If this
    point is the same as the start point, then a good closed chain has been
    found, so set the closed flag accordingly, then exit.  */

       *(buffer + *current_point) = 0;        /* Blank out the detected point */

       if (*current_point == chain -> start)
	 return( chain -> closed = 1);
    }
  }

/*............................................................................
	Chain code an entire image, calculating various parameters.

        This routine chain codes the entire image into several chains.
	It is passed two pointers to image buffers, the first contains the
	image to chain code, the other to be used as a work buffer.  Also
	a pointer to the output chain storage area is passed, along with
	the size of the storage area in bytes.
	To start with a copy of the input image is made in the other buffer.
	This copy is then untraced as the chain is coded.  For each closed
        chain found, the area, perimeter and shape factor are calculated.
*/

chain_code (PIXEL           *buffer,
            PIXEL           *work_buffer,
	    char   	    *work_area,
	    unsigned int    *work_area_size)
	    /*
            int    *chains_found,
	    ) */
  {
    int            chain_attempts,
		   backup_mode,
		   backup_attempts,
		   chain_nr = 0;

    unsigned int   start_position,
		   current_point,
		   length_save,
		   bytes_left;

    CHAIN          work_chain[1];
    char           ch, *buff, *next_free_byte;

    next_free_byte = work_area;
    bytes_left = *work_area_size;

    if ((work_chain -> element = farmalloc ( 10000)) == NULL) {
      printf( "\7\nUnable to allocate buffer storage");
      return;
    }

    display_status( "Chain coding image" );

/* Create a copy of the image in the other buffer */

    memcpy( work_buffer, buffer, 65535 );
/*    store_buffer( work_buffer );
    freeze();  */

    start_position = 0;

    for( chain_attempts = 0; chain_attempts < 500; chain_attempts++ ) {
      if( scan_start_point (work_buffer, &start_position )) {
	work_chain -> start = start_position;
	for (backup_attempts = 0; backup_attempts < 5; backup_attempts++) {
  	  calc_chain ( work_buffer, work_chain, backup_attempts, &current_point );
	  if (work_chain -> closed)
	    break;
	  if (!backup_attempts)
	    length_save = work_chain -> length;
          if ((work_chain -> length--) < 1) {
	    work_chain -> length = length_save;
	    break;
	  }
          switch ( *(work_chain -> element + work_chain -> length)) {
	    case 0 : current_point -= 257;
	             break;
	    case 1 : current_point -= 256;
		     break;
	    case 2 : current_point -= 255;
		     break;
	    case 3 : current_point += 1;
		     break;
	    case 4 : current_point += 257;
		     break;
	    case 5 : current_point += 256;
		     break;
	    case 6 : current_point += 255;
		     break;
	    case 7 : current_point -= 1;
		     break;
	    default: printf("\nBad chain detected !");
		     break;
	  }
	}
        if ((work_chain -> length > 5) && (work_chain -> closed)) {
	  if( work_chain -> length < bytes_left && chain_nr < NO_OF_CHAINS ) {
  	    chains[chain_nr].length = work_chain -> length;
	    chains[chain_nr].start = work_chain -> start;
	    chains[chain_nr].closed = work_chain -> closed;
	    chains[chain_nr].element = next_free_byte;
	    memcpy( next_free_byte, work_chain -> element, work_chain -> length );
	    bytes_left -= work_chain -> length;
	    next_free_byte += work_chain -> length;
	    chain_nr++;
	  }
/*          if ((buff = malloc ((screen_width + 1) * (screen_height + 1) * 2)) == NULL) {
            printf("\7\nUnable to allocate necessary memory");
            exit( 1 );
          }
          gettext( 1, 1, screen_width, screen_height, buff );
          display_chain( work_chain );
          puttext( 1, 1, screen_width, screen_height, buff );
          free( buff );  */
        }
      }
    }
    farfree( work_chain -> element );
  }


/*...........................................................................
	Display a chain vector into a buffer

     Required	: chain		Chain vector to display
*/
display_chain_to_buffer( PIXEL *buffer, CHAIN *chain)
  {
    int	i, temp;

   buffer += chain -> start;

/* Now trace the outline of the chain */

    temp = chain -> length + 2;
    for ( i = 0; (i < temp ) && (i < 10000); i++ ) {
      *buffer = 255;
      switch ( *(chain -> element + i )) {
   	case 0 : buffer -= 257;
		 break;
   	case 1 : buffer -= 256;
		 break;
   	case 2 : buffer -= 255;
		 break;
   	case 3 : buffer++;
		 break;
   	case 4 : buffer += 257;
		 break;
   	case 5 : buffer += 256;
		 break;
   	case 6 : buffer += 255;
		 break;
   	case 7 : buffer--;
		 break;
      }
    }
  }



/*...........................................................................
	Display a chain vector on a MGA screen

     Required	: chain		Chain vector to display
*/
display_chain (CHAIN *chain)
  {
    int x, x_start, x_max,
	y, y_start, y_max,
	i, temp, dot, g_error,
	graph_width = 512,
	graph_height = 256;

    struct palettetype palette;

    int graphdriver = DETECT, graphmode;
    char ch;

    graphic = 1;

    initgraph( &graphdriver, &graphmode, "");
    if( graphdriver < 0 ) {
      display_status( grapherrormsg( g_error = graphresult() ) );
      }

    getpalette( &palette );
    dot = palette.size - 1;

    x_start = ( getmaxx() - graph_width ) / 2;
    y_start = ( getmaxy() - graph_height ) / 2;
    x_max = x_start + graph_width;
    y_max = y_start + graph_height;
    if( y_start < 0 )  {      		/* For CGA */
      y_start = 0;
      y_max = 200;
    }

/* Calculate the start co-ordinates */

    y = chain -> start / 256 + y_start;
    x = ( chain -> start % 256 ) * 2 + x_start;

/* Draw the box outline */

    line( x_start, y_max, x_max, y_max );
    line( x_start, y_start, x_max, y_start );
    line( x_start, y_start, x_start, y_max );
    line( x_max, y_start, x_max, y_max );

/* Now trace the outline of the chain */

    temp = chain -> length + 2;
    for ( i = 0; (i < temp ) && (i < 10000); i++ ) {
      putpixel ( x, y, dot );
      putpixel ( x + 1, y, dot );
      switch ( *(chain -> element + i )) {
   	case 0 : x -= 2;
		 y--;
		 break;
   	case 1 : y--;
		 break;
   	case 2 : x += 2;
		 y--;
		 break;
   	case 3 : x += 2;
		 break;
   	case 4 : x += 2;
		 y++;
		 break;
   	case 5 : y++;
		 break;
   	case 6 : x -= 2;
		 y++;
		 break;
   	case 7 : x -= 2;
		 break;
      }
    }
    cscanf ("%c", &ch );
    closegraph();
    graphic = 0;
  }

camera (PIXEL *buffer, MENU *main_menu_params)
  {
    MENU menu_params;

    menu_params.x = main_menu_params -> x + main_menu_params -> width;
    menu_params.y = main_menu_params -> y + main_menu_params -> opt + 1;
    menu_params.width = 20;
    menu_params.opt = 2;
    menu_params.optmax = 3;
    menu_params.new = 1;

    do {
      if ( !menu_disp ( &menu_params,
             "Camera Functions",
	     "Free run",
	     "Freeze",
	     "Load Image"))     return(0);

      switch ( menu_params.opt )
	  {
	    case 1:
	      free_run ();
	      break;
	    case 2:
	      freeze ();
	      break;
	    case 3:
	      load_buffer (buffer);
	      freeze();
	      break;
	    default:
	      cprintf ("\7");
	      break;
          }
    }
    while (1);
  }

disk( PIXEL *buffer, MENU *main_menu_params)
  {
    MENU menu_params;

    menu_params.x = main_menu_params -> x + main_menu_params -> width;
    menu_params.y = main_menu_params -> y + main_menu_params -> opt + 1;
    menu_params.width = 16;
    menu_params.opt = 2;
    menu_params.optmax = 2;
    menu_params.new = 1;

    do {
      if( !menu_disp( &menu_params,
             "Disk Functions",
	     "Save image",
	     "Retrive image"))  {
	return(0);
      }

      switch( menu_params.opt )
	  {
	    case 1:
	      buffer_to_disk (buffer, &menu_params);
	      break;
	    case 2:
	      if( disk_to_buffer (buffer, &menu_params) == 0 ) {
                store_buffer (buffer);
	        freeze();
	      }
	      break;
	    default:
	      cprintf ("\7");
	      break;
          }
    }
    while (1);
  }

process (PIXEL *in_buffer, PIXEL *out_buffer, MENU *main_menu_params)
  {
    MENU            menu_params;
    unsigned long   l;
    unsigned int    work_area_size;
    char            *work_area;
    int             i;

    menu_params.x = main_menu_params -> x + main_menu_params -> width;
    menu_params.y = main_menu_params -> y + main_menu_params -> opt + 1;
    menu_params.width = 22;
    menu_params.opt = 1;
    menu_params.optmax = 4;
    menu_params.new = 1;

    do {
      if ( !menu_disp ( &menu_params,
             "Processing Functions",
	     "Equalise",
	     "Histogram",
	     "Edge detect",
	     "Chain code")) {
         set_result_window();
	 clear_result_window();
	 return(0);
       }

      switch( menu_params.opt )
	  {
	    case 1:
	      equalise( in_buffer, out_buffer );
              store_buffer( out_buffer );
              freeze();
              memcpy( in_buffer, out_buffer, 65535 );
	      break;
	    case 2:
	      histogram( in_buffer, out_buffer );
              store_buffer( out_buffer);
              freeze();
	      break;
	    case 3:
	      roberts_grad( in_buffer, out_buffer );
              store_buffer( out_buffer );
              freeze();
              memcpy( in_buffer, out_buffer, 65535 );
	      break;
	    case 4:
	      for( i = 0; i < NO_OF_CHAINS; i++, chains[i].length = 0 );

	      if(( l = coreleft() - 10000 ) > MAX_CHAIN_BUFFER_SIZE ) {
		l = MAX_CHAIN_BUFFER_SIZE;
	      }
	      work_area_size = l;
	      if( work_area_size < 10000 ) {
                printf("\7\nUnable to allocate buffer storage");
                return( 1 );
              }
	      work_area = malloc( work_area_size );
	      chain_code( in_buffer, out_buffer, work_area, &work_area_size );

              memset( (char*)out_buffer, 0, 65535 );

	      set_result_window();
	      if( chains[0].length == 0 ) {
		gotoxy( 10, wherey() );
		cprintf( "\nNo chains found" );
	      }
	      else {
	        for( i = 0; chains[i].length != 0; i++ ) {
  	          gotoxy( 10, wherey() );
   	          cprintf( "\nChain %d %d long, starts %u stored %p",
			 i,
			 chains[i].length,
			 chains[i].start,
			 chains[i].element );
		  delay( 100 );
		  display_chain_to_buffer( out_buffer, &chains[i] );
	        }
	      }
              free( work_area );
              clear_result_window();
	      store_buffer( out_buffer );
	      freeze();
	      break;
	    default:
	      cprintf ("\7");
	      break;
          }
    }
    while (1);
  }

options (MENU *main_menu_params)
  {
    MENU menu_params;

    menu_params.x = main_menu_params -> x + main_menu_params -> width;
    menu_params.y = main_menu_params -> y + main_menu_params -> opt + 1;
    menu_params.width = 22;
    menu_params.opt = 1;
    menu_params.optmax = 3;
    menu_params.new = 1;

    do {
      if ( !menu_disp ( &menu_params,
             "Setup Options",
	     "Quadrant/Single mode",
	     "Graphic display",
	     "Test memory"))  {
 	set_result_window();
	clear_result_window();
        return(0);
      }
      switch ( menu_params.opt )
	  {
	    case 1:
	      if (!quadrant) {
	        if ((quad_buf_start = (PIXEL *) farmalloc(65570 * sizeof(PIXEL))) == NULL) {
                printf("\7\nUnable to allocate buffer storage");
                return( 1 );
	        }
              quadrant = 1;
	      buffer9_start = quad_buf_start;
	      buffer9_start = MK_FP(FP_SEG(buffer9_start) + 1, 0);
	      memset ( buffer9_start, 0, 65535);
	      }
	      else {
  	        farfree(quad_buf_start);
                quadrant = 0;		/* Set to single display mode */
	      }
	      display_status( "" );
	      break;
	    case 2:
	      if( !graphic_disp ) {
		graphic_disp = 1;
	      }
	      else {
                graphic_disp = 0;	/* Set to single display mode */
	      }
	      display_status( "" );
	      break;
	    case 3:
	      test_memory();
	      break;
	    default:
	      cprintf ("\7");
	      break;
          }
    }
    while (1);
  }

draw_screen()
  {
    int i;

    clrscr();
    putchar( BOXLT );
    for( i = 2; i < screen_width; i++ )
      putchar( BOXHB );
    putchar( BOXRT );
    for( i = 2; i < screen_height; i++ ) {
      gotoxy( 1, i );
      putchar( BOXVB );
      gotoxy( screen_width, i );
      putchar( BOXVB );
    }
    gotoxy( 1, screen_height );
    putchar( BOXLB );
    for( i = 2; i < screen_width; i++ )
      putchar( BOXHB );
    putchar( BOXRB );

    gotoxy( 1, screen_height - 2 );
    putchar( BOXLJX );
    for( i = 2; i < screen_width; i++ )
      putchar( BOXHBX );
    putchar( BOXRJX );

    gotoxy( 1, 3 );
    putchar( BOXLJ );
    for( i = 2; i < screen_width; i++ )
      putchar( BOXHB );
    putchar( BOXRJ );

  }


main ()
  {
    MENU  menu_params;

    graphic = 0;		/* Set graphics flag off */
    graphic_disp = 0;		/* Set not to use graphics mode */
    quadrant = 0;		/* Set to single display mode */

    draw_screen();
    gotoxy (20,2);
    highvideo();
    cprintf("F R A M E   G R A B B E R    D E M O");
    normvideo();

    if ((((buffer1_start = (PIXEL *) farmalloc(65536 * sizeof(PIXEL))) == NULL)) |
       (((buffer2_start = (PIXEL *) farmalloc(65536 * sizeof(PIXEL))) == NULL))) {
      printf("\7\nUnable to allocate buffer storage");
      exit(1);
    }

    free_run ();

    menu_params.x = 10;
    menu_params.y = 5;
    menu_params.width = 10;
    menu_params.opt = 1;
    menu_params.optmax = 4;
    menu_params.new = 1;

    do {
      if ( !menu_disp ( &menu_params,
             "M E N U",
	     "Camera",
	     "Disk",
	     "Process",
	     "Options") ) {
	       clrscr();
	       exit(1);
      }
      switch ( menu_params.opt )
	  {
	    case 1:
	      camera (buffer1_start, &menu_params);
	      break;
	    case 2:
	      disk (buffer1_start, &menu_params);
	      break;
	    case 3:
	      process (buffer1_start, buffer2_start, &menu_params);
	      break;
	    case 4:
	      options (&menu_params);
	      break;
	    default:
	      cprintf ("\7");
	      break;
          }
    }
    while (1);
  }
